#!/usr/bin/env python3

import numpy as np


def splice_imgs(imgs):
    npc = np.concatenate
    if len(imgs) == 1:
        return imgs[0]
    if len(imgs) == 2:
        return npc([imgs[0], imgs[1]], 0)
    if len(imgs) >= 3:
        idx = 0
        while idx < len(imgs):
            if not idx:
                big = npc([imgs[0], imgs[1]], 1)
            else:
                if len(imgs) - idx == 1:
                    big = npc([big[::2, ::2], imgs[idx]], 0)
                else:
                    big = npc([big, npc([imgs[idx], imgs[idx + 1]], 1)], 0)[::2, ::2]
            idx += 2
        return big


def merge_depth_image_for_vis(data):
    color = data["color"]
    depth = data["depth"]
    from skimage.exposure import equalize_hist

    vis = (
        color / 255 * (1 - equalize_hist(np.where(depth == 0, 80930, depth)))[..., None]
    )
    return (vis * 255).astype(np.uint8)


if __name__ == "__main__":
    pass



#!/usr/bin/env python3

import cv2
import time
import numpy as np
import pyrealsense2 as rs
from threading import Timer

try:
    from class_as_process import class_as_process
    from rs_utils import splice_imgs, merge_depth_image_for_vis
except ModuleNotFoundError:
    from .class_as_process import class_as_process
    from .rs_utils import splice_imgs, merge_depth_image_for_vis


class RealsenseCamera:
    def __init__(self, snid, get_frame_method="wait_for_frames", try_num=-1):
        self.snid = snid
        self.get_frame_method = get_frame_method
        self.try_num = try_num
        time.sleep(0.1 * np.random.random())
        self.try_run(self.set_config_and_option)

    def robust_get_data(self):
        """may_has_delay"""
        return self.try_run(self.get_data)

    def get_data(self):
        frameset = self.get_frameset()
        data = self.post_processing(frameset)
        return data

    def get_frameset(self):
        get_frameset = getattr(self.pipeline, self.get_frame_method)
        frameset = get_frameset()
        return frameset

    def set_config_and_option(self):
        # set config
        config = rs.config()
        config.enable_stream(rs.stream.depth, 1280, 720, rs.format.z16, 6)
        config.enable_stream(rs.stream.color, 1280, 720, rs.format.rgb8, 6)
        config.enable_device(self.snid)

        # set pipline
        self.pipeline = rs.pipeline()
        self.profile = self.pipeline.start(config)
        self.device = self.profile.get_device()

        # set depth option
        depth_sensor = self.depth_sensor = self.device.first_depth_sensor()

        # set "High Accuracy" depth mode
        depth_sensor.set_option(rs.option.visual_preset, 3)

        # set_laser_power
        power_rate = 1
        depth_sensor.set_option(
            rs.option.laser_power,
            int(depth_sensor.get_option_range(rs.option.laser_power).max * power_rate),
        )

    align = rs.align(rs.stream.color)
    spatial = rs.spatial_filter()
    temporal = rs.temporal_filter()

    def post_processing(self, frameset):
        streams = self.profile.get_streams()
        assert frameset.size() == len(streams)
        data = {}
        frameset = self.align.process(frameset)
        for stream in streams:
            key = str(stream.stream_type()).replace("stream.", "")
            if rs.stream.infrared == stream.stream_type():
                frame = frameset.get_infrared_frame(stream.stream_index())
                key = key + str(stream.stream_index())
            else:
                frame = frameset.first_or_default(stream.stream_type())
                if stream.stream_type() == rs.stream.depth:
                    frame = self.spatial.process(frame)
                    frame = self.temporal.process(frame)
            data[key] = self.frame_to_numpy(frame)
        return data

    @staticmethod
    def frame_to_numpy(frame):
        return np.asanyarray(frame.get_data())

    def try_run(
        self, funcation,
    ):
        n = 0
        while True:
            try:
                n += 1
                return funcation()
            except RuntimeError as e:
                if n == self.try_num:
                    print(
                        f"[snid:{self.snid}] try {funcation} {self.try_num} times, fails!"
                    )
                    raise e
                print(f"[snid:{self.snid}] try {n} time, RuntimeError:")
                print(e)
                # self.hardware_reset()
                try:
                    if "device" not in self.__dict__:
                        devices = list(rs.context().devices)
                        filter_devices = [d for d in devices if self.snid in str(d)]
                        assert (
                            len(filter_devices) == 1
                        ), f"S/N: {self.snid} not in {devices}"
                        self.device = filter_devices[0]
                    self.device.hardware_reset()
                    time.sleep(1.5)
                    """
                    1.5s is a magic number
                    if sleep(1), will need 5s at `pipe.start(config)` after `device.hardware_reset()`
                    """
                except RuntimeError as e:
                    print(f"[snid:{self.snid}] try {n} time, RuntimeError:")
                    print(e)
                    if "No such file or directory" in str(e):
                        pass
                try:
                    self.set_config_and_option()
                except RuntimeError as e:
                    print(f"[snid:{self.snid}] try {n} time, RuntimeError:")
                    print(e)

    def hardware_reset(self):
        """
        To del
        """
        self.device.hardware_reset()
        time.sleep(1.5)
        self.set_config_and_option()

    @classmethod
    def from_camera_idx(cls, idx=0):
        snids = MultiRealsenseManger.get_all_snids()
        snid = snids[idx]
        return cls(snid)

    @staticmethod
    def get_all_snids(context=None):
        if context is None:
            context = rs.context()
        snids = []
        for d in context.devices:
            if "realsense" in d.get_info(rs.camera_info.name).lower():
                snids.append(d.get_info(rs.camera_info.serial_number))
        return sorted(snids)

    def load_settings_json(self, json_path):
        with open(json_path, "r") as file:
            json_text = file.read().strip()
        advanced_mode = rs.rs400_advanced_mode(self.device)
        advanced_mode.load_json(json_text)

    def save_settings_json(self, json_path="/tmp/realsense-settings.json"):
        advanced_mode = rs.rs400_advanced_mode(self.device)
        json_text = advanced_mode.serialize_json()
        with open(json_path, "w") as file:
            file.write(json_text)

    def stop(self):
        self.pipeline.stop()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.stop()


class RealsenseForVis(RealsenseCamera):
    def post_processing(self, frameset):
        data = super().post_processing(frameset)
        # data["vis"] = data["color"]
        data["vis"] = merge_depth_image_for_vis(data)
        return data


class MultiRealsenseManger(dict):
    def __init__(
        self,
        MetaRealsenseClass=RealsenseCamera,
        init_args=None,
        init_kwargs=None,
        dummy=False,
        snids=None,
    ):
        """
        Multi-realsense manger with multi-processing
        based on dict

        Parameters
        ----------
        MetaRealsenseClass : TYPE, optional
            Realsense class. The default is RealsenseCamera.
        init_args : tuple, optional
            Init args for MetaRealsenseClass. The default is [].
        init_kwargs : dict, optional
            Init kwargs for MetaRealsenseClass. The default is {}.
        dummy : bool, optional
            If True, using multiprocess.dummy which is multi-threading. 
            The default is False.
        snids : list of str, optional
            List of snids. The default is self.get_all_snids().
        """

        super().__init__()
        if snids is None:
            snids = MultiRealsenseManger.get_all_snids()
        for snid in snids:
            _kwargs = {} if init_kwargs is None else init_kwargs.copy()
            _kwargs.update(snid=snid)
            self[snid] = class_as_process(
                cls=MetaRealsenseClass,
                init_args=init_args,
                init_kwargs=_kwargs,
                dummy=dummy,
            )

    get_all_snids = RealsenseCamera.get_all_snids

    def __enter__(self):
        return self

    def __exit__(self, *args):
        [process.terminate() for process in self.values()]
        [process.join() for process in self.values()]

    def __getattr__(self, key):
        if not callable(getattr(next(iter(self.values())), key)):
            return {k: getattr(v, key) for k, v in self.items()}

        def func(*args, **kwargs):
            threads = []
            res = {}

            def func(snid, process):
                res[snid] = getattr(process, key)(*args, **kwargs)

            for snid, process in self.items():
                thread = Timer(0, func, (snid, process))
                thread.start()
                threads.append(thread)
            [thread.join() for thread in threads]
            res = {key: res[key] for key in sorted(res)}
            return res

        return func


if __name__ == "__main__":
    """
    Robust multi-processing multi-realsense manager
    
    Usage:
        1. Build a new class based on RealsenseCamera.
        2. Override set_config_and_option and post_processing method
            for custom config and processing.
    
    Example:
        >>> class CustomRealsense(RealsenseCamera):
                def set_config_and_option(self):
                    # edit the config and option 
                    # refer to RealsenseCamera.set_config_and_option
                
                def post_processing(self, frameset):
                    # edit the post processing for frameset
                    # refer to RealsenseCamera.post_processing
                    return data        
    """

    # Example code
    snids = MultiRealsenseManger.get_all_snids()
    print("All realsense snids:", snids)
    MetaRealsenseClass = RealsenseForVis
    FRAME_NUM = 50
    with MetaRealsenseClass(snids[0]) as rc:
        print(f"imshow first realsense's image, {FRAME_NUM} frames")
        for i in range(FRAME_NUM):
            print(f"{i}/{FRAME_NUM}")
            cv2.waitKey(1)
            data = rc.robust_get_data()
            vis = merge_depth_image_for_vis(data)
            cv2.imshow("first camera", vis[..., ::-1])

    with MultiRealsenseManger(MetaRealsenseClass) as mrm:
        for snid in mrm:
            for i in range(FRAME_NUM):
                print(f"snid: {snid}, {i}/{FRAME_NUM}")
                cv2.waitKey(1)
                data = mrm[snid].robust_get_data()
                vis = merge_depth_image_for_vis(data)
                cv2.imshow(f"snid: {snid}", vis[..., ::-1])

        print("\n\n\n imshow multi realsense images")
        while 1:
            _begin = time.time()
            key = cv2.waitKey(1)
            datas = mrm.robust_get_data()
            frame = splice_imgs([data["vis"] for data in datas.values()])
            cv2.imshow(f"{len(mrm)} realsense", frame[..., ::-1])
            if key == ord("q"):
                cv2.destroyAllWindows()
                break
            print("Loop Spend", round(time.time() - _begin, 2))



__title__ = "multi_realsense_manager"
__version__ = "0.1.7"
__author__ = "DIYer22"
__author_email__ = "ylxx@live.com"
__url__ = "https://github.com/DIYer22/multi_realsense_manager"
__description__ = "Robust high-performance multi-realsenses manager"



#!/usr/bin/env python3


def class_as_process(cls, init_args=None, init_kwargs=None, dummy=False):
    if dummy:
        from multiprocessing.dummy import Process, Pipe
    else:
        from multiprocessing import Process, Pipe

    class ClassAsProcess(Process):
        def __init__(self, cls, init_args=None, init_kwargs=None):
            if init_args is None:
                init_args = []
            if init_kwargs is None:
                init_kwargs = {}
            self._dirs = set(dir(self))
            self.conn, conn_child = Pipe()
            super().__init__(
                target=self.new_process, args=(cls, init_args, init_kwargs, conn_child)
            )
            self.start()

        @staticmethod
        def new_process(cls, init_args, init_kwargs, conn):
            instance = cls(*init_args, **init_kwargs)

            def listen():
                while True:
                    rev = conn.recv()
                    if rev is None:
                        break
                    if isinstance(rev, dict):
                        if "method" in rev:
                            func = getattr(instance, rev["method"])
                            re = func(*rev.get("args", []), **rev.get("kwargs", {}))
                            conn.send(re)

            if getattr(instance, "__enter__", None) and getattr(
                instance, "__exit__", None
            ):
                with instance:
                    listen()
            else:
                listen()

        def terminate(self):
            self.conn.send(None)

        def __getattr__(self, key):
            def func(*args, **kwargs):
                self.conn.send(dict(method=key, args=args, kwargs=kwargs))
                return self.conn.recv()

            return func

    return ClassAsProcess(cls, init_args, init_kwargs)


if __name__ == "__main__":
    # test
    class A:
        def __enter__(self):
            print("__enter__")
            return self

        def __exit__(self, *args):
            print("__exit__")

        def test(self, *args, **kwargs):
            return f"call args: {args} kwargs: {kwargs}"

    a = class_as_process(A, dummy=0)
    print(a.test(1, 2, a="a", b="b"))
    a.terminate()
    a.join()
